from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("reelname")
except PackageNotFoundError:  # pragma: no cover
    # Fallback for local dev or editable installs
    __version__ = "0.0.0"



import re
from re import Pattern

# Matches titles with a bracketed year like:
#   "Title (2023)", "Title [2023]", "Title {2023}", "Title <2023>"
# Regex breakdown:
#   (?P<title>.+?)        → Named group "title": matches any character (non-greedy)
#   \s*                   → Optional whitespace between title and year
#   [\(\[\{\<]            → Match any opening bracket: (, [, {, or <
#   (?P<year>\d{4})       → Named group "year": exactly 4 digits
#   [\)\]\}\>]            → Match any closing bracket: ), ], }, or >
BRACKETED_PATTERN: Pattern[str] = re.compile(
    r"(?P<title>.+?)\s*[\(\[\{\<](?P<year>\d{4})[\)\]\}\>]"
)

# Matches titles with a dot-year-dot pattern:
#   "Title.2023." or "Some.Show.2010."
# Regex breakdown:
#   ^                    → Anchor to start of string
#   (?P<title>.+?)       → Named group "title": any characters (non-greedy), up to the year
#   \.                   → Literal dot before the year
#   (?P<year>\d{4})      → Named group "year": exactly 4 digits
#   \.                   → Literal dot after the year
DOT_YEAR_PATTERN: Pattern[str] = re.compile(r"(?P<title>.+?)\.(?P<year>\d{4})\.")

# Matches titles with space-separated year:
#   "Title 2023" or "Title 2023 Something"
# Regex breakdown:
#   ^                      → Anchor to start of string
#   (?P<title>.+?)         → Named group "title": any characters (non-greedy), up to year
#   \s+                    → One or more spaces
#   (?P<year>\d{4})        → Named group "year": exactly 4 digits
#   (?:\s|$)               → Non-capturing group: either a space or end of string
SPACE_YEAR_PATTERN: Pattern[str] = re.compile(r"(?P<title>.+?)\s+(?P<year>\d{4})(?:\s|$)")

# Matches characters that are not allowed in Windows filenames:
#   < > : " / \ | ? *
# Regex breakdown:
#   [<>:"/\\|?*]           → Match any one of these characters literally
#   Double escaping \\     → Needed for backslash inside regex string
#   strip spaces and . at the end
INVALID_FILENAME_CHARS: Pattern[str] = re.compile(r'[<>:"/\\|?*]|[ .]+$')

# Matches tracker prefixes at the beginning of filenames like:
#   "www.1TamilMV.fi - ", "yts.mx - ", "1337x.to - "
# Regex breakdown:
#   ^                            → Anchor to start of string
#   (?:                          → Non-capturing group:
#       www\.[\w.-]+             → Starts with "www." followed by domain characters
#       (letters, digits, dash, dot)
#       |                        → OR
#       \w+\.\w{2,}              → Domain name like "yts.mx" or "1337x.to"
#   )
#   (?:\s*[--]\s*)               → Optional spaces, then dash or en dash (two hyphens),
#   then optional spaces
URL_PREFIX_PATTERN: Pattern[str] = re.compile(
    r"^(?:www\.[\w.-]+|\w+\.\w{2,})(?:\s*[--]\s*)", re.IGNORECASE
)



import pytest

from reelname.utils import extract_title_and_year, rebuild_filename

from .data import EXTRACT_CASES, MOVIE_RENAME_CASES


@pytest.mark.parametrize("filename, expected", EXTRACT_CASES)
def test_extract_title_and_year(filename: str, expected: str) -> None:
    assert extract_title_and_year(filename) == expected


@pytest.mark.parametrize("orig, expected_new", MOVIE_RENAME_CASES)
def test_rebuild_filename_matches_extract(orig: str, expected_new: str) -> None:
    """
    For each movie style filename, extract title+year, then rebuild and
    compare to the expected renamed form.
    """
    # derive title/year from extract_title_and_year
    title, year = extract_title_and_year(orig)
    rebuilt = rebuild_filename(orig, title, year)  # type: ignore
    assert rebuilt == expected_new



from __future__ import annotations

from pathlib import Path
import re

import aiofiles.os
import click
from imdb import Cinemagoer
from rapidfuzz import fuzz

from .constants import (
    BRACKETED_PATTERN,
    DOT_YEAR_PATTERN,
    INVALID_FILENAME_CHARS,
    SPACE_YEAR_PATTERN,
    URL_PREFIX_PATTERN,
)


def extract_title_and_year(filename: str) -> tuple[str | None, str | None]:
    """
    Extract (title, year) from a filename. Handles:
      - optional site/tracker prefixes (e.g. 'www.site.com - ')
      - bracketed years (e.g. [2023], (2022))
      - dot-year (e.g. Title.2023.)
      - space-year (e.g. Title 2023 ...)
    """
    # Drop a tracker prefix if present
    if m := URL_PREFIX_PATTERN.match(filename):
        filename = filename[m.end() :]  # remove only prefix, not all ' - '

    # Try year-extracting patterns
    for pat in (BRACKETED_PATTERN, DOT_YEAR_PATTERN, SPACE_YEAR_PATTERN):
        if m := pat.search(filename):
            raw_title = m.group("title").strip()
            year = m.group("year")

            # Normalize dot/underscore separators to spaces
            title = re.sub(r"[._]+", " ", raw_title).strip()
            return title, year

    return None, None


def get_match_score(title: str, candidate: str) -> float:
    """
    Combine three metrics and take the minimum:
      - fuzz.ratio
      - fuzz.token_sort_ratio
      - fuzz.partial_token_sort_ratio
    This punishes missing tokens, order-changes, and trivial substrings.
    """
    return min(
        fuzz.ratio(title, candidate),
        fuzz.token_sort_ratio(title, candidate),
        fuzz.partial_token_sort_ratio(title, candidate),
    )


def fetch_info_from_imdb(title: str, year: str) -> tuple[str, str]:
    """
    Look up the title on IMDb via Cinemagoer:
      - Search IMDb for the raw title.
      - Use fuzzy matching to pick the best.
      - If found, return (official_title, imdb_year).
      - Otherwise, fall back to the extracted (title, year).
    """
    ia = Cinemagoer()
    results = ia.search_movie(f"{title} {year}")

    best_match = None
    best_score = 0.0
    for movie in results:
        candidate = movie.get("title")
        if not candidate:
            continue
        score = get_match_score(title, candidate)
        if score > best_score:
            movie_year = movie.get("year")
            if not movie_year:
                ia.update(movie)  # fetch full details for this movie
            if str(movie_year) != year:
                continue
            best_score = score
            best_match = movie
            if score >= 98.0:
                break

    if best_match and best_score >= 80:
        imdb_title = best_match.get("title")
        imdb_year = best_match.get("year")
        if imdb_title and imdb_year:  # return the fixed title and year
            click.echo(f"🔎 Found: {imdb_title} ({imdb_year})")
            return imdb_title, str(imdb_year)

    return title, year


def _sanitize_filename(name: str) -> str:
    """
    Remove any characters that are invalid in filenames on Windows (and many other OSes).
    """
    return INVALID_FILENAME_CHARS.sub("", name).strip()


def rebuild_filename(original: str, title: str, year: str) -> str:
    """
    Given a cleaned filename starting at the title/year, reconstruct it as:
      {title} ({year}){suffix}

    Always uses parentheses around the year, and strips any leftover
    punctuation or brackets before the suffix.
    """
    # 1) Find the first occurrence of the year in the string
    idx = original.find(year)
    if idx == -1:
        # (shouldn't happen if caller extracted title/year correctly)
        file_name = f"{title} ({year})"
    else:
        # 2) Take everything after that year
        suffix = original[idx + len(year) :]

        # 3) Strip any leading punctuation/brackets/whitespace from the suffix,
        #    replacing it with a single space (if there is any suffix at all).
        suffix_clean = re.sub(r"^[\s._\-()\[\]{}<>]+", " ", suffix)

        # 4) Build the new filename
        file_name = f"{title} ({year}){suffix_clean}"

    return _sanitize_filename(file_name)


async def _rename_file_async(old_path: str, new_path: str) -> None:
    await aiofiles.os.rename(old_path, new_path)


async def _process_files(directory: Path) -> None:
    """
    Scan `directory` for files, skip ones without title/year,
    skip already-formatted ones, lookup IMDb, and rename.
    """
    for file in directory.iterdir():
        if not file.is_file():
            continue

        raw_filename = file.name

        # 1) extract_title_and_year now handles prefix-stripping
        title, year = extract_title_and_year(raw_filename)
        if not title or not year:
            click.echo(f"⏩ Skipping (no title/year): {raw_filename}")
            continue

        # 2) skip files already beginning with “Title (Year)”
        formatted_prefix = f"{title} ({year})"
        if raw_filename.startswith(formatted_prefix):
            click.echo(f"⏩ Skipping already formatted: {raw_filename}")
            continue

        click.echo(f"🔎 Looking up: {title} ({year})")
        imdb_title, imdb_year = fetch_info_from_imdb(title, year)

        if not imdb_title or not imdb_year:
            click.echo(f"⏩ Skipping (no IMDb match): {raw_filename}")
            continue

        # 3) rebuild_filename expects the *cleaned* substring:
        #    we re-slice from the first occurrence of the year
        #    so that rebuild_filename sees “Title(…)suffix”
        start = raw_filename.find(year)
        cleaned = raw_filename[start:]

        new_name = rebuild_filename(cleaned, imdb_title, imdb_year)
        if new_name != raw_filename:
            # rename the *original* file to the cleaned new_name
            await _rename_file_async(str(file), str(file.parent / new_name))
            click.echo(f"✅ Renamed: {raw_filename} → {new_name}")
        else:
            click.echo(f"⏩ Already correct: {raw_filename}")



#!/usr/bin/env python3
"""
Check release date for a given version in CHANGELOG.md.
"""

from datetime import date
from pathlib import Path
import sys

CHANGELOG_PATH = Path("CHANGELOG.md")


def main(version: str) -> None:
    if not CHANGELOG_PATH.exists():
        print(f"❌ ERROR: {CHANGELOG_PATH} does not exist.")
        sys.exit(1)

    today = date.today().isoformat()
    target_line = f"## [{version}] - {today}"

    found = False
    lines = CHANGELOG_PATH.read_text(encoding="utf-8").splitlines()

    for i, line in enumerate(lines):
        if line.strip() == target_line:
            print(f"🔍 Found line {i + 1}: {target_line}")
            found = True
            break

    if not found:
        print("❌ ERROR: CHANGELOG.md is not ready for release.")
        print(f"   Expected line: {target_line}")
        print("Tip: Check if it's still marked as '[Unreleased]' and update it to today's date.")
        sys.exit(1)


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: uv run python scripts/check_changelog_date.py <version>")
        sys.exit(1)

    main(sys.argv[1])



import asyncio
from pathlib import Path

import click
from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer

from .utils import _process_files


class FileChangeHandler(FileSystemEventHandler):
    """Trigger _process_files when new files appear."""

    def __init__(self, directory: Path, loop: asyncio.AbstractEventLoop) -> None:
        super().__init__()
        self.directory = directory
        self.loop = loop

    def on_created(self, event: FileSystemEvent) -> None:
        if not event.is_directory:
            click.echo(f"🔔 Detected new file: {event.src_path!s}")
            asyncio.run_coroutine_threadsafe(_process_files(self.directory), self.loop)


def watch_directory(directory: str) -> None:
    """
    Continuously watch `directory` and perform IMDb-based renaming.
    """
    dir_path = Path(directory)
    loop = asyncio.get_event_loop()
    handler = FileChangeHandler(dir_path, loop)
    observer = Observer()
    observer.schedule(handler, str(dir_path), recursive=False)
    observer.start()
    click.echo(f"👀 Watching: {dir_path!s}")

    try:
        loop.run_forever()
    except KeyboardInterrupt:
        click.echo("👋 Stopping watcher...")
        observer.stop()
    observer.join()


def run_once(directory: str) -> None:
    """
    Perform a single-pass scan of `directory` and rename matching files.
    """
    dir_path = Path(directory)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(_process_files(dir_path))
    click.echo("✅ Run-once processing complete.")



# Files we expect to rename, plus their expected new names
MOVIE_RENAME_CASES = [
    (
        "Inception.2010.1080p.BluRay.x264-REF.mkv",
        "Inception (2010) 1080p.BluRay.x264-REF.mkv",
    ),
    (
        "The.Dark.Knight.2008.720p.WEB-DL.DD5.1.H.264-FLAC-TK.mkv",
        "The Dark Knight (2008) 720p.WEB-DL.DD5.1.H.264-FLAC-TK.mkv",
    ),
    (
        "Parasite.2019.1080p.WEB-DL.DD5.1.H.264-ABCD.mkv",
        "Parasite (2019) 1080p.WEB-DL.DD5.1.H.264-ABCD.mkv",
    ),
    (
        "The.Studio.2025.S01E07.Casting.1080p.ATVP.WEB-DL.DDP5.1.Atmos.H.264-FLUX.mkv",
        "The Studio (2025) S01E07.Casting.1080p.ATVP.WEB-DL.DDP5.1.Atmos.H.264-FLUX.mkv",
    ),
    (
        "The.Dark.Knight[2008]DvDrip-aXXo.avi",
        "The Dark Knight (2008) DvDrip-aXXo.avi",
    ),
    (
        "Anchorman.2.The.Legend.Continues.2013.1080p.BluRay.DDP.5.1.H.265-EDGE2020.mkv",
        "Anchorman 2 The Legend Continues (2013) 1080p.BluRay.DDP.5.1.H.265-EDGE2020.mkv",
    ),
    (
        "The Straight Story 1999 REMASTERED 1080p BluRay HEVC x265 5.1 BONE.mkv",
        "The Straight Story (1999) REMASTERED 1080p BluRay HEVC x265 5.1 BONE.mkv",
    ),
    (
        "www.1TamilMV.fi - Bajirao Mastani (2015) Tamil HQ HDRip - 1080p - HEVC - x265 - (AAC 2.0) - 1.8GB - ESub.mkv",  # noqa: E501
        "Bajirao Mastani (2015) Tamil HQ HDRip - 1080p - HEVC - x265 - (AAC 2.0) - 1.8GB - ESub.mkv",  # noqa: E501
    ),
    (
        "Spider-Man.2002.1080p.BluRay.x264-GROUP.mkv",
        "Spider-Man (2002) 1080p.BluRay.x264-GROUP.mkv",
    ),
    (
        "Inception<2010>720p.x264.AAC.mkv",
        "Inception (2010) 720p.x264.AAC.mkv",
    ),
    (
        "Interstellar{2014}1080p.BluRay.x265-PSA.mkv",
        "Interstellar (2014) 1080p.BluRay.x265-PSA.mkv",
    ),
    (
        "The.Matrix.1999.REMASTERED.BluRay.1080p.DTS-HD.MA.5.1.mkv",
        "The Matrix (1999) REMASTERED.BluRay.1080p.DTS-HD.MA.5.1.mkv",
    ),
    (
        "yts.mx - Anora (2024) 1080p BluRay x265 HEVC-YTS.mp4",
        "Anora (2024) 1080p BluRay x265 HEVC-YTS.mp4",
    ),
    (
        "www.YTS.mx - The.Dark.Knight.2008.1080p.BluRay.x264.mkv",
        "The Dark Knight (2008) 1080p.BluRay.x264.mkv",
    ),
    (
        "rarbg.to - The Matrix 1999 1080p BluRay x264 AAC.mkv",
        "The Matrix (1999) 1080p BluRay x264 AAC.mkv",
    ),
]

# Filenames that should be skipped because they lack a parsable year
SKIP_CASES = [
    "Some.Random.Show.S02E05.HDTV-XYZ.mkv",
    "JustAFileWithNoYear.txt",
    "LOL. Body.of.Proof.S02E15.HDTV.XviD-LOL.mkv",
    "ASAP. One.Tree.Hill.S09E07.HDTV.x264-ASAP.mkv",
    "2HD. Suburgatory.S01E15.READNFO.HDTV.XviD-2HD.mkv",
    "LOL. The.Mentalist.S04E16.HDTV.x264-LOL.mkv",
]

# Cases for extract_title_and_year (both positive and negative)
EXTRACT_CASES = [
    ("Kanchana (2011) Tamil TRUE WEB-DL.mkv", ("Kanchana", "2011")),
    ("The.Studio.2025.S01E07.Casting.mkv", ("The Studio", "2025")),
    ("NoYearHere.mkv", (None, None)),
    ("Citizenfour.2014.720p.WEB-DL.AAC2.0.H.264-NOGRP.mp4", ("Citizenfour", "2014")),
    ("The.Movie.Title.2010.REMASTERED.1080p.BluRay.x264-GROUP.mkv", ("The Movie Title", "2010")),
    ("Title.Of.The.Movie.2023.BluRay.x264-GROUP.mkv", ("Title Of The Movie", "2023")),
    (
        "The.Studio.2025.S01E07.Casting.1080p.ATVP.WEB-DL.DDP5.1.Atmos.H.264-FLUX.mkv",
        ("The Studio", "2025"),
    ),
    (
        "www.1TamilMV.fi - Bajirao Mastani (2015) Tamil HQ HDRip - 1080p - HEVC - x265 - (AAC 2.0) - 1.8GB - ESub.mkv",  # noqa: E501
        ("Bajirao Mastani", "2015"),
    ),
    ("The.Dark.Knight[2008]DvDrip-aXXo.avi", ("The Dark Knight", "2008")),
    ("www.1TamilMV.fi - Bajirao Mastani (2015) Tamil HQ HDRip.mkv", ("Bajirao Mastani", "2015")),
    ("Spider-Man.2002.1080p.BluRay.x264-GROUP.mkv", ("Spider-Man", "2002")),
    ("Inception<2010>720p.x264.AAC.mkv", ("Inception", "2010")),
    ("Interstellar{2014}1080p.BluRay.x265-PSA.mkv", ("Interstellar", "2014")),
    ("The.Matrix.1999.REMASTERED.BluRay.1080p.DTS-HD.MA.5.1.mkv", ("The Matrix", "1999")),
    ("yts.mx - Dune (2021) 1080p BluRay x265 HEVC-YTS.mp4", ("Dune", "2021")),
    # Case where dot-year appears mid-string after prefix
    ("www.YTS.mx - The.Dark.Knight.2008.1080p.BluRay.x264.mkv", ("The Dark Knight", "2008")),
    # Case where space-year appears mid-string after prefix
    ("rarbg.to - The Matrix 1999 1080p BluRay x264 AAC.mkv", ("The Matrix", "1999")),
] + [(fn, (None, None)) for fn in SKIP_CASES]



from pathlib import Path

from click.testing import CliRunner
import pytest

from reelname import __version__, cli
from reelname.cli import main

from .data import MOVIE_RENAME_CASES, SKIP_CASES


@pytest.mark.parametrize("options", [["-h"], ["--help"]])
def test_help(options: list[str]) -> None:
    runner = CliRunner()
    result = runner.invoke(cli.main, options)

    assert result.exit_code == 0, f"Help options {options} failed with exit code {result.exit_code}"
    assert result.output.startswith(
        "Usage: "
    ), f"Help output for {options} did not start with 'Usage:'"
    assert "-h, --help" in result.output, f"Help flags missing from output for {options}"


@pytest.mark.parametrize("options", [["-v"], ["--version"]])
def test_version(options: list[str]) -> None:
    runner = CliRunner()
    result = runner.invoke(cli.main, options)

    assert (
        result.exit_code == 0
    ), f"Version options {options} failed with exit code {result.exit_code}"
    assert (
        __version__ in result.output
    ), f"Expected version {__version__} not found in output for {options}"


@pytest.mark.parametrize("orig, expected", MOVIE_RENAME_CASES)
def test_real_run_once_renames(tmp_path: Path, orig: str, expected: str) -> None:
    # create the sample file
    src = tmp_path / orig
    src.write_bytes(b"")

    runner = CliRunner()
    result = runner.invoke(main, [str(tmp_path)])
    assert result.exit_code == 0, result.output

    # original removed, expected exists
    assert not src.exists(), f"Original still exists: {orig}"
    dst = tmp_path / expected
    assert dst.exists(), f"Expected renamed file not found: {expected}"
    assert f"✅ Renamed: {orig} → {expected}" in result.output


@pytest.mark.parametrize("orig", SKIP_CASES)
def test_real_run_once_skips(tmp_path: Path, orig: str) -> None:
    src = tmp_path / orig
    src.write_bytes(b"")

    runner = CliRunner()
    result = runner.invoke(main, [str(tmp_path)])
    assert result.exit_code == 0, result.output

    assert src.exists(), f"File was unexpectedly renamed: {orig}"
    assert "Skipping" in result.output



import click

from ._version import __version__
from .reelname import run_once, watch_directory


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(__version__, "-v", "--version")
@click.option(
    "-w",
    "--watch",
    is_flag=True,
    default=False,
    help="Continuously watch directory instead of one-time run",
)
@click.argument(
    "directory",
    metavar="<directory>",
    nargs=1,
    required=False,
    type=click.Path(exists=True, file_okay=False),
)
def main(directory: str, watch: bool) -> None:
    """
    Rename media files by correcting their titles via IMDb.

    By default runs once; use -w/--watch to run continuously.

    \b
    Examples:
      reelname /path/to/media
      reelname -w /path/to/media
    """
    if not directory:
        directory = "."
    if watch:
        watch_directory(directory)
    else:
        run_once(directory)



from __future__ import annotations


class ReelNameError(Exception):
    """
    Base exception. All other exceptions
    inherit from here.
    """

    detail = "An error occurred."

    def __init__(self, extra_detail: str | None = None) -> None:
        super().__init__()
        self.extra_detail = extra_detail

    def __str__(self) -> str:
        if self.extra_detail:
            return f"{self.detail} :: {self.extra_detail}"
        return self.detail
